// Copyright 2015 Noitom Technology Co., Ltd. All Rights Reserved.

#include "BVHAssetDetailsCustomization.h"
#include "Developer/DesktopWidgets/Public/Widgets/Input/SFilePathPicker.h"
#include "Widgets/Images/SImage.h"

#define LOCTEXT_NAMESPACE "BVHAssetDetailsCustomization"


/* IDetailCustomization interface
 *****************************************************************************/
BEGIN_SLATE_FUNCTION_BUILD_OPTIMIZATION
void FBVHAssetDetailsCustomization::CustomizeDetails(IDetailLayoutBuilder& DetailBuilder)
{
	DetailBuilder.GetObjectsBeingCustomized(CustomizedBVHAssets);

	// customize 'File' category
	IDetailCategoryBuilder& SourceCategory = DetailBuilder.EditCategory("File");
	{
		// BVHFilePath
		FilePathProperty = DetailBuilder.GetProperty("BVHFilePath");
		{
			IDetailPropertyRow& FilePathRow = SourceCategory.AddProperty(FilePathProperty);

			FilePathRow.DisplayName(LOCTEXT("BVHFilePath", "BVH File Path"));
			FilePathRow.CustomWidget()
				.NameContent()
				[
					SNew(SHorizontalBox)

					+ SHorizontalBox::Slot()
						.AutoWidth()
						.VAlign(VAlign_Center)
						[
							SNew(STextBlock)
								.Font(IDetailLayoutBuilder::GetDetailFont())
								.Text(LOCTEXT("BVHFilePath", "BVH File Path"))
								.ToolTipText(LOCTEXT("BVHFilePath", "BVH File Path"))
						]

					+ SHorizontalBox::Slot()
						.FillWidth(1.0f)
						.HAlign(HAlign_Left)
						.VAlign(VAlign_Center)
						.Padding(4.0f, 0.0f, 0.0f, 0.0f)
						[
							SNew(SImage)
								.Image(FCoreStyle::Get().GetBrush("Icons.Warning"))
								.ToolTipText(LOCTEXT("InvalidFilePathWarning", "The current path points to a file that does not exist or is not located inside the /Content/BVH/ directory."))
								.Visibility(this, &FBVHAssetDetailsCustomization::HandleFilePathWarningIconVisibility)
						]
				]
				.ValueContent()
				.MaxDesiredWidth(0.0f)
				.MinDesiredWidth(125.0f)
				[
					SNew(SFilePathPicker)
						.BrowseButtonImage(FEditorStyle::GetBrush("PropertyWindow.Button_Ellipsis"))
						.BrowseButtonStyle(FEditorStyle::Get(), "HoverHintOnly")
						.BrowseButtonToolTip(LOCTEXT("BVHFileBrowseButtonToolTip", "Choose a *.bvh file from this computer."))
#if (ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION < 18)
						.BrowseDirectory(FPaths::GameContentDir() / TEXT("BVH"))
#else
						.BrowseDirectory(FPaths::ProjectContentDir() / TEXT("BVH"))
#endif
						.FilePath(this, &FBVHAssetDetailsCustomization::HandleBVHFilePickerFilePath)
						.FileTypeFilter(this, &FBVHAssetDetailsCustomization::HandleBVHFilePickerFileTypeFilter)
						.OnPathPicked(this, &FBVHAssetDetailsCustomization::HandleBVHFilePickerPathPicked)
						.ToolTipText(LOCTEXT("BVHFilePathToolTip", "The path to a BVH file on this computer."))
				];
		}
	}
}
END_SLATE_FUNCTION_BUILD_OPTIMIZATION

/* FBVHAssetDetailsCustomization callbacks
 *****************************************************************************/
FString FBVHAssetDetailsCustomization::HandleBVHFilePickerFilePath() const
{
	FString FilePath;
	FilePathProperty->GetValue(FilePath);
	return FilePath;
}

FString FBVHAssetDetailsCustomization::HandleBVHFilePickerFileTypeFilter() const
{
	FString Filter = TEXT("All files (*.*)|*.*");
	Filter = Filter + TEXT("|BVH File (*.bvh)|*.bvh");
	return Filter;
}

void FBVHAssetDetailsCustomization::HandleBVHFilePickerPathPicked(const FString& PickedPath)
{
	if (PickedPath.IsEmpty() || PickedPath.StartsWith(TEXT("./")) || PickedPath.Contains(TEXT("://")))
	{
		FilePathProperty->SetValue(PickedPath);
	}
	else
	{	
		FString FullBVHFilePath = FPaths::ConvertRelativePathToFull(PickedPath);
#if (ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION < 18)
		const FString FullGameContentDir = FPaths::ConvertRelativePathToFull(FPaths::GameContentDir());
#else
		const FString FullGameContentDir = FPaths::ConvertRelativePathToFull (FPaths::ProjectContentDir ());
#endif

		if (FullBVHFilePath.StartsWith(FullGameContentDir))
		{
			FPaths::MakePathRelativeTo(FullBVHFilePath, *FullGameContentDir);
			FullBVHFilePath = FString(TEXT("./")) + FullBVHFilePath;
		}

		FilePathProperty->SetValue(FullBVHFilePath);
	}
}

EVisibility FBVHAssetDetailsCustomization::HandleFilePathWarningIconVisibility() const
{
	FString FilePath;

	if ((FilePathProperty->GetValue(FilePath) != FPropertyAccess::Success) || FilePath.IsEmpty() || FilePath.Contains(TEXT("://")))
	{
		return EVisibility::Hidden;
	}

#if (ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION < 18)
	const FString FullBVHDirPath = FPaths::ConvertRelativePathToFull(FPaths::GameContentDir() / TEXT("BVH"));
	const FString FullBVHFilePath = FPaths::ConvertRelativePathToFull(FPaths::IsRelative(FilePath) ? FPaths::GameContentDir() / FilePath : FilePath);
#else
	const FString FullBVHDirPath = FPaths::ConvertRelativePathToFull(FPaths::ProjectContentDir() / TEXT("BVH"));
	const FString FullBVHFilePath = FPaths::ConvertRelativePathToFull(FPaths::IsRelative(FilePath) ? FPaths::ProjectContentDir() / FilePath : FilePath);
#endif

	if (FullBVHFilePath.StartsWith(FullBVHDirPath))
	{
		if (FPaths::FileExists(FullBVHFilePath))
		{
			return EVisibility::Hidden;
		}

		// file doesn't exist
		return EVisibility::Visible;
	}

	// file not inside BVH folder
	return EVisibility::Visible;
}


#undef LOCTEXT_NAMESPACE
