// Copyright 2015 Noitom Technology Co., Ltd. All Rights Reserved.

#include "AnimationDataManager.h"
#include "MotionCaptureHelper.h"
#include "Hi5PacketParser.h"
#include "BVHAsset.h"
#include "MotionCaptureAvatarComponent.h"
#include "PluginLog.h"

namespace Hi5
{
	using ELogVerbosity::Verbose;

	FAnimationDataPtr FAnimationDataManager::InvalidPtr;
	FAnimationDataManager* FAnimationDataManager::sInstance = nullptr;

	FAnimationDataManager* FAnimationDataManager::GetInstance( )
	{
		if (!sInstance)
		{
			sInstance = new FAnimationDataManager( );
		}

		return sInstance;
	}

	void FAnimationDataManager::AddMotionCaptureComponent( UMotionCaptureAvatarComponent* component )
	{
		MotionCaptureComponentList.Add( component );
	}

	void FAnimationDataManager::RemoveMotionCaptureComponent( UMotionCaptureAvatarComponent* component )
	{
		MotionCaptureComponentList.Remove( component );
	}

#if WITH_AVATARID
	bool FAnimationDataManager::AddAnimationData( int32 AvatarID, const FAnimationDataPtr& NewAnimData )
	{
		if (!NewAnimData.IsValid( ) || (INDEX_NONE == AvatarID))
		{
			return false;
		}

		PN_LOG( Verbose, TEXT( "Add new animation data ID:%d." ), AvatarID );

		FPNWriteScopeLock Lock( &ReadWriteLock );
		CharAnimDatas.Add( { AvatarID }, NewAnimData );

		return true;
	}

	bool FAnimationDataManager::UpdateAnimationData( int32 AvatarID, const FAnimationDataPtr& NewAnimData )
	{
		if ((0 == CharAnimDatas.Num( )) || (INDEX_NONE == AvatarID))
		{
			return false;
		}

		PN_LOG( Verbose, TEXT( "Get animation data ID:%d." ), AvatarID );

		FPNWriteScopeLock Lock( &ReadWriteLock );
		FAnimDBuffer* DataBuffer = CharAnimDatas.Find( { AvatarID } );
		if (nullptr == DataBuffer)
		{
			return false;
		}

		DataBuffer->Set( NewAnimData );
		DataBuffer->SwapBuffer( );

		return true;
	}

	bool FAnimationDataManager::GetAnimationData( int32 AvatarID, FAnimationData& OutAnimData )
	{
		if ((0 == CharAnimDatas.Num( )) || (INDEX_NONE == AvatarID))
		{
			return false;
		}

		PN_LOG( Verbose, TEXT( "Get animation data ID:%d." ), AvatarID );

		FAnimDBuffer* DataBuffer = CharAnimDatas.Find( { AvatarID } );
		if (nullptr == DataBuffer)
		{
			return false;
		}

		OutAnimData = *(DataBuffer->Get( ));

		return true;
	}

	bool FAnimationDataManager::IsAnimationDataExist( int32 AvatarID )
	{
		if ((0 == CharAnimDatas.Num( )) || (INDEX_NONE == AvatarID))
		{
			return false;
		}

		FPNReadScopeLock Lock( &ReadWriteLock );
		const FAnimDBuffer* DataBuffer = CharAnimDatas.Find( { AvatarID } );
		if (nullptr == DataBuffer)
		{
			return false;
		}
		return true;
	}

	bool FAnimationDataManager::RemoveAnimationData( int32 AvatarID )
	{
		// 	if ((0 == CharAnimDatas.Num()) || (INDEX_NONE == AvatarID))
		// 	{
		// 		return false;
		// 	}
		// 
		// 	PN_LOG(Verbose, TEXT("Remove animation data ID:%d."), AvatarID);
		// 
		// 	FPNWriteScopeLock Lock(&ReadWriteLock);
		// 	CharAnimDatas.FindAndRemoveChecked({ AvatarID });
		// 
		// 	return true;

		return false;
	}
#else
	bool FAnimationDataManager::AddAnimationData( const FName& AvatarName, FAnimationDataPtr& NewAnimData )
	{
		if (!NewAnimData.IsValid( ) || (AvatarName == NAME_None))
		{
			return false;
		}

		PN_LOG( Verbose, TEXT( "Add new animation data Name:%s." ), *(AvatarName.ToString( )) );

		FPNWriteScopeLock Lock( &ReadWriteLock );
		CharAnimDatas.Add( { AvatarName }, NewAnimData );

		return true;
	}

	bool FAnimationDataManager::UpdateAnimationData( const FName& AvatarName, FAnimationDataPtr& NewAnimData )
	{
		if ((0 == CharAnimDatas.Num( )) || (AvatarName == NAME_None))
		{
			return false;
		}

		PN_LOG( Verbose, TEXT( "Get animation data Name:%s." ), *(AvatarName.ToString( )) );

		FPNWriteScopeLock Lock( &ReadWriteLock );
		FAnimDBuffer* DataBuffer = CharAnimDatas.Find( { AvatarName } );
		if (nullptr == DataBuffer)
		{
			return false;
		}

		DataBuffer->Set( NewAnimData );
		DataBuffer->SwapBuffer( );

		return true;
	}

	bool FAnimationDataManager::GetAnimationData( const FName& AvatarName, FAnimationData& OutAnimData )
	{
		if ((0 == CharAnimDatas.Num( )) || (AvatarName == NAME_None))
		{
			return false;
		}

		PN_LOG( Verbose, TEXT( "Get animation data Name:%s." ), *(AvatarName.ToString( )) );

		FPNReadScopeLock Lock( &ReadWriteLock );
		FAnimDBuffer* DataBuffer = CharAnimDatas.Find( { AvatarName } );
		if (nullptr == DataBuffer)
		{
			return false;
		}

		OutAnimData = *(DataBuffer->Get( ));

		return true;
	}

	bool FAnimationDataManager::IsAnimationDataExist( const FName& AvatarName )
	{
		if ((0 == CharAnimDatas.Num( )) || (AvatarName == NAME_None))
		{
			return false;
		}

		FPNReadScopeLock Lock( &ReadWriteLock );
		const FAnimDBuffer* DataBuffer = CharAnimDatas.Find( { AvatarName } );
		if (nullptr == DataBuffer)
		{
			return false;
		}
		return true;
	}


	bool FAnimationDataManager::AddHandAnimationData( const FName& AvatarName, FAnimationDataPtr& NewAnimData, int32 HandMode )
	{
		if (!NewAnimData.IsValid( ) || (AvatarName == NAME_None))
		{
			return false;
		}

		FPNWriteScopeLock Lock( &ReadWriteLock );
		if (HandMode == 1)
		{
			LeftHandAnimDatas.Add( { AvatarName }, NewAnimData );
			PN_LOG( Verbose, TEXT( "Add new left hand animation data Name: %s." ), *(AvatarName.ToString( )) );
		}
		else if (HandMode == 2)
		{
			RightHandAnimDatas.Add( { AvatarName }, NewAnimData );
			PN_LOG( Verbose, TEXT( "Add new right hand animation data Name: %s." ), *(AvatarName.ToString( )) );
		}
		else
		{
			return false;
		}

		return true;
	}

	bool FAnimationDataManager::UpdateHandAnimationData( const FName& AvatarName, FAnimationDataPtr& NewAnimData, int32 HandMode )
	{
		if (HandMode == 1)
		{
			if ((0 == LeftHandAnimDatas.Num( )) || (AvatarName == NAME_None))
			{
				return false;
			}
		}
		else if (HandMode == 2)
		{
			if ((0 == RightHandAnimDatas.Num( )) || (AvatarName == NAME_None))
			{
				return false;
			}
		}

		FPNWriteScopeLock Lock( &ReadWriteLock );
		FAnimDBuffer* DataBuffer = nullptr;
		if (HandMode == 1)
		{
			DataBuffer = LeftHandAnimDatas.Find( { AvatarName } );
			PN_LOG( Verbose, TEXT( "Get left hand animation data Name: %s." ), *(AvatarName.ToString( )) );
		}
		else if (HandMode == 2)
		{
			DataBuffer = RightHandAnimDatas.Find( { AvatarName } );
			PN_LOG( Verbose, TEXT( "Get right hand animation data Name: %s." ), *(AvatarName.ToString( )) );
		}
		else
		{
			return false;
		}

		if (nullptr == DataBuffer)
		{
			return false;
		}

		DataBuffer->Set( NewAnimData );
		DataBuffer->SwapBuffer( );

		return true;
	}

	bool FAnimationDataManager::GetHandAnimationData( const FName& AvatarName, FAnimationData& OutAnimData, int32 HandMode )
	{
		if (HandMode == 1)
		{
			if ((0 == LeftHandAnimDatas.Num( )) || (AvatarName == NAME_None))
			{
				return false;
			}
		}
		else if (HandMode == 2)
		{
			if ((0 == RightHandAnimDatas.Num( )) || (AvatarName == NAME_None))
			{
				return false;
			}
		}
		else
		{
			return false;
		}

		FPNReadScopeLock Lock( &ReadWriteLock );
		FAnimDBuffer* DataBuffer = nullptr;
		if (HandMode == 1)
		{
			DataBuffer = LeftHandAnimDatas.Find( { AvatarName } );
			PN_LOG( Verbose, TEXT( "Get left hand animation data Name: %s." ), *(AvatarName.ToString( )) );
		}
		else if (HandMode == 2)
		{
			DataBuffer = RightHandAnimDatas.Find( { AvatarName } );
			PN_LOG( Verbose, TEXT( "Get right hand animation data Name: %s." ), *(AvatarName.ToString( )) );
		}
		else
		{
			return false;
		}

		if (nullptr == DataBuffer)
		{
			return false;
		}

		OutAnimData = *(DataBuffer->Get( ));

		return true;
	}

	bool FAnimationDataManager::IsHandAnimationDataExist( const FName& AvatarName, int32 HandMode )
	{
		if (HandMode == 1)
		{
			if ((0 == LeftHandAnimDatas.Num( )) || (AvatarName == NAME_None))
			{
				return false;
			}
			else
			{
				FPNReadScopeLock Lock( &ReadWriteLock );
				const FAnimDBuffer* DataBuffer = LeftHandAnimDatas.Find( { AvatarName } );
				if (nullptr == DataBuffer)
				{
					return false;
				}
				return true;
			}
		}
		else if (HandMode == 2)
		{
			if ((0 == RightHandAnimDatas.Num( )) || (AvatarName == NAME_None))
			{
				return false;
			}
			else
			{
				FPNReadScopeLock Lock( &ReadWriteLock );
				const FAnimDBuffer* DataBuffer = RightHandAnimDatas.Find( { AvatarName } );
				if (nullptr == DataBuffer)
				{
					return false;
				}
				return true;
			}
		}
		else
		{
			return false;
		}
	}



	bool FAnimationDataManager::RemoveAnimationData( const FName& AvatarName )
	{
		// 	if ((0 == CharAnimDatas.Num()) || (AvatarName == NAME_None))
		// 	{
		// 		return false;
		// 	}
		// 
		// 	PN_LOG(Verbose, TEXT("Remove animation data Name:%s."), *(AvatarName.ToString()));
		// 
		// 	FPNWriteScopeLock Lock(&ReadWriteLock);
		// 	CharAnimDatas.FindAndRemoveChecked({ AvatarName });
		// 
		// 	return true;

		return false;
	}
#endif

	void FAnimationDataManager::RemoveAllAnimationData( )
	{
		FPNWriteScopeLock Lock( &ReadWriteLock );
		CharAnimDatas.Empty( );
	}

	bool FAnimationDataManager::GetAnimationData( const UBVHAsset* BVHAsset, double& OutFramePointer, FAnimationData& OutAnimData )
	{
		OutAnimData.bUseRootEffect = BVHAsset->bUseRootEffect;
		OutAnimData.bWithDisplacement = BVHAsset->bWithDisplacement;

		if (OutFramePointer > BVHAsset->FrameTime * (BVHAsset->Frames - 1))
		{
			OutFramePointer = 0.0;
		}

		double Step = OutFramePointer / BVHAsset->FrameTime;
		int32 FrameUp = FMath::FloorToInt( Step ) + BVHAsset->FrameDataOffset;
		int32 FrameDown = 1 + FrameUp;

		float BlendFactor = (float)(Step - FMath::FloorToInt( Step ));

		FString Data0 = BVHAsset->AnimData[FrameUp];
		TArray<FString> Words0;
		Data0.ParseIntoArray( Words0, TEXT( " " ), false );

		FString Data1 = BVHAsset->AnimData[FrameDown];
		TArray<FString> Words1;
		Data1.ParseIntoArray( Words1, TEXT( " " ), false );

		for (int32 i = 0; i < BVHAsset->BoneTotalNumber; ++i)
		{
			int32 Index = OutAnimData.bWithDisplacement ? i * 6 : (EMCBones::RightHand == i ? 0 : i * 3 + 3);
			uint8 PNBoneIndex = FMotionCaptureHelper::PNBoneIndices[i];

			int32 Flag = 0;
			FVector PosUp = FVector::ZeroVector;
			if ((EMCBones::RightHand == i) || OutAnimData.bWithDisplacement)
			{
				float X = FCString::Atof( *Words0[Index + Flag++] );
				float Y = -FCString::Atof( *Words0[Index + Flag++] );
				float Z = FCString::Atof( *Words0[Index + Flag++] );
				/*PosUp = FVector(X, Y, Z);*/ // zbjxb
				PosUp = FVector( Z, -X, -Y );
			}
			float QX0 = FCString::Atof( *Words0[Index + Flag++] );
			float QY0 = FCString::Atof( *Words0[Index + Flag++] );
			float QZ0 = FCString::Atof( *Words0[Index + Flag++] );
			FQuat QuaUp = FMotionCaptureHelper::GetQuaternion( QX0, QY0, QZ0 );

			Flag = 0;
			FVector PosDown = FVector::ZeroVector;
			if ((EMCBones::RightHand == i) || OutAnimData.bWithDisplacement)
			{
				float X = FCString::Atof( *Words1[Index + Flag++] );
				float Y = -FCString::Atof( *Words1[Index + Flag++] );
				float Z = FCString::Atof( *Words1[Index + Flag++] );
				/*PosDown = FVector(X, Y, Z);*/ // zbjxb
				PosDown = FVector( Z, -X, -Y );
			}
			float QX1 = FCString::Atof( *Words1[Index + Flag++] );
			float QY1 = FCString::Atof( *Words1[Index + Flag++] );
			float QZ1 = FCString::Atof( *Words1[Index + Flag++] );
			FQuat QuaDown = FMotionCaptureHelper::GetQuaternion( QX1, QY1, QZ1 );

			if ((EMCBones::RightHand == i) || OutAnimData.bWithDisplacement)
			{
				OutAnimData.BonePositions[PNBoneIndex] = PosUp * BlendFactor + PosDown * (1 - BlendFactor);
			}

			FQuat SlerpQuat = FQuat::Slerp( QuaDown, QuaUp, BlendFactor );
			SlerpQuat.Normalize( );
			OutAnimData.BoneRotations[PNBoneIndex] = SlerpQuat;
		}

		// temp useless
		// OutAnimData.RootPosition = ;
		// OutAnimData.RootRotation = ;

		return true;
	}

	bool FAnimationDataManager::IsBVHFileNameValid( const FString& BVHFileName ) const
	{
		if (FString( TEXT( "" ) ) == BVHFileName)
		{
			return false;
		}

		if (0 != FPaths::GetExtension( BVHFileName ).Compare( TEXT( "bvh" ), ESearchCase::IgnoreCase ))
		{
			return false;
		}

		return true;
	}

	void FAnimationDataManager::BeginRead( )
	{
		ReadWriteLock.LockRead( );
	}

	void FAnimationDataManager::EndRead( )
	{
		ReadWriteLock.UnlockRead( );
	}

	FName FAnimationDataManager::GetLocalAvatarName( )
	{
		if (MotionCaptureComponentList.Num( ) <= 0)
		{
			return NAME_None;
		}

		for (int index = 0; index < MotionCaptureComponentList.Num( ); ++index)
		{
			if (MotionCaptureComponentList[index]->IsLocalAvatar)
			{
				return MotionCaptureComponentList[index]->AvatarName;
			}
		}

		return NAME_None;
	}

	bool FAnimationDataManager::IsAvatarValid( const FName& InAvatarName, UObject** PawnObj )
	{
		*PawnObj = nullptr;

		if (MotionCaptureComponentList.Num( ) <= 0)
		{
			return false;
		}

		for (int index = 0; index < MotionCaptureComponentList.Num( ); ++index)
		{
			if (MotionCaptureComponentList[index]->AvatarName.IsEqual( InAvatarName ))
			{
				*PawnObj = Cast<UObject>( MotionCaptureComponentList[index]->GetOwner( ) );
				return true;
			}
		}

		return false;
	}
}
