// Copyright 2015 Noitom Technology Co., Ltd. All Rights Reserved.

#pragma once

#include "CoreMinimal.h"
#include "AnimationData.h"
#include "PNThreadingBase.h"


class UBVHAsset;
class UMotionCaptureAvatarComponent;

namespace Hi5
{
	class FAnimationData;

	class FAnimDBuffer
	{
	public:
		FAnimDBuffer( FAnimationDataPtr& NewAnimData )
		{
			Data[0] = MakeShareable( new FAnimationData( ) );
			//Data[1] = MakeShareable(new FAnimationData());

			Swap( Data[0], NewAnimData );
		}

		const FAnimationDataPtr& Get( ) const
		{
			return Data[Index];
		}

		void Set( FAnimationDataPtr& NewAnimData )
		{
			Swap( Data[Index/*(Index + 1) % 2*/], NewAnimData );
		}

		// Not thread safety, because only network thread can call SwapBuffer
		void SwapBuffer( )
		{
			//Index = ++Index % 2;
		}

	private:
		volatile uint8 Index = 0;

		// The reason why store pointer but not object is for avoid copy consumption.
		FAnimationDataPtr Data[2];
	};


	class FAnimDataIndexKey
	{
		friend struct FAnimDataMapKeyFuncs;

	public:
#if WITH_AVATARID
		FAnimDataIndexKey( int32 InID ) :
			ID( InID )
		{
		}
#else
		FAnimDataIndexKey( const FName& InName ) :
			Name( InName )
		{
		}
#endif

#if WITH_AVATARID
		int32 ID = INDEX_NONE;
#else
		FName Name = NAME_None;
#endif

	private:
		uint32 KeyHash = (uint32)INDEX_NONE;
	};

	struct FAnimDataMapKeyFuncs : BaseKeyFuncs<FAnimDBuffer, FAnimDataIndexKey>
	{
		FORCEINLINE static const FAnimDataIndexKey& GetSetKey( const TPair<FAnimDataIndexKey, FAnimDBuffer>& Element )
		{
			return Element.Key;
		}
		FORCEINLINE static bool Matches( const FAnimDataIndexKey& A, const FAnimDataIndexKey& B )
		{
#if WITH_AVATARID
			if (A.ID == B.ID)
			{
				return true;
			}
#else
			if (A.Name == B.Name)
			{
				return true;
			}
#endif
			return false;
		}
		FORCEINLINE static uint32 GetKeyHash( const FAnimDataIndexKey& Key )
		{
			return Key.KeyHash;
		}
	};


	class FAnimationDataManager
	{
		friend class FADMReadScopeLock;

	public:
		static FAnimationDataManager* GetInstance( );

		~FAnimationDataManager( ) = delete;

#if WITH_AVATARID
		bool AddAnimationData( int32 AvatarID, const FAnimationDataPtr& NewAnimData );

		bool UpdateAnimationData( int32 AvatarID, const FAnimationDataPtr& NewAnimData );

		bool GetAnimationData( int32 AvatarID, FAnimationData& OutAnimData );

		bool IsAnimationDataExist( int32 AvatarID );

		bool RemoveAnimationData( int32 AvatarID );
#else
		bool AddAnimationData( const FName& AvatarName, FAnimationDataPtr& NewAnimData );

		bool UpdateAnimationData( const FName& AvatarName, FAnimationDataPtr& NewAnimData );

		bool GetAnimationData( const FName& AvatarName, FAnimationData& OutAnimData );

		bool IsAnimationDataExist( const FName& AvatarName );

		bool RemoveAnimationData( const FName& AvatarName );
#endif
		// Don`t use this at present
		void RemoveAllAnimationData( );

		bool GetAnimationData( const UBVHAsset* BVHAsset, double& OutFramePointer, FAnimationData& OutAnimData );

		void AddMotionCaptureComponent( UMotionCaptureAvatarComponent* component );
		void RemoveMotionCaptureComponent( UMotionCaptureAvatarComponent* component );

		FName GetLocalAvatarName( );
		bool IsAvatarValid( const FName& InAvatarName, UObject** PawnObj );



		// 手部数据
		bool AddHandAnimationData( const FName& AvatarName, FAnimationDataPtr& NewAnimData, int32 HandMode );

		bool UpdateHandAnimationData( const FName& AvatarName, FAnimationDataPtr& NewAnimData, int32 HandMode );

		bool GetHandAnimationData( const FName& AvatarName, FAnimationData& OutAnimData, int32 HandMode );

		bool IsHandAnimationDataExist( const FName& AvatarName, int32 HandMode );


	private:
		bool IsBVHFileNameValid( const FString& BVHFileName ) const;

		void BeginRead( );
		void EndRead( );

		TMap<FAnimDataIndexKey, FAnimDBuffer, FDefaultSetAllocator, FAnimDataMapKeyFuncs> CharAnimDatas;
		TMap<FAnimDataIndexKey, FAnimDBuffer, FDefaultSetAllocator, FAnimDataMapKeyFuncs> LeftHandAnimDatas;
		TMap<FAnimDataIndexKey, FAnimDBuffer, FDefaultSetAllocator, FAnimDataMapKeyFuncs> RightHandAnimDatas;

		FPNMultiReaderSingleWriter ReadWriteLock;

		static FAnimationDataPtr InvalidPtr;
		static FAnimationDataManager* sInstance;

		TArray<UMotionCaptureAvatarComponent*> MotionCaptureComponentList;
	};


	class FADMReadScopeLock
	{
	public:
		FADMReadScopeLock( )
		{
			FAnimationDataManager::GetInstance( )->BeginRead( );
		}

		~FADMReadScopeLock( )
		{
			FAnimationDataManager::GetInstance( )->EndRead( );
		}

		FADMReadScopeLock( const FADMReadScopeLock& InScopeLock ) = delete;

		FADMReadScopeLock& operator=( const FADMReadScopeLock& InScopeLock ) = delete;
	};

}
