﻿
#include "HI5Glove.h"
#include <iostream>
#include <thread>
#include <vector>
#include <algorithm>
#include "NetworkManager.h"
#include "AnimationDataManager.h"
#include "ISteamVRPlugin.h"
//#include <future>
#include "GameDelegates.h"
#include "IMotionCaptureModule.h"
#include "Hi5_Manager_Thread.h"
#if USE_ALICE
IHybridData& AliceIns () {
	return *IHybridData::GetPtr ();
}
#endif

std::function<bool(float*, float*)> LeftCallback;
std::function<bool(float*, float*)> RightCallback;

bool bCalirating = false;
std::thread* CalibrateThread = nullptr;

class ThreadController
{
public:
	ThreadController ()
	{

	}

	~ThreadController () {
		waitForFinished ();	// block if the func_wrapper still keep running
	}

	template<typename _Fn,
		class... _Args,
		class = typename std::enable_if<
			!std::is_same<typename std::decay<_Fn>::type, std::thread>::value>::type>
	void run(_Fn&& func, _Args&&... args)
	{
		m_promise.Reset (new TPromise<void> ());
		m_future = m_promise->GetFuture ();

		auto func_wrapper = std::bind (std::forward<_Fn> (func), std::forward<_Args> (args)...);
		auto f = [=]()
		{
			//modify by zhouxin 2017-8-23
			//try {
				func_wrapper ();
				this->m_promise->SetValue ();
			/*}
			catch(...)
			{
				this->m_promise->set_exception (std::current_exception ());
			}*/
		};
		m_worker = std::thread (f);
	}

	void waitForFinished ()
	{
		if (m_promise)
		{
			if (m_worker.joinable ())
			{
				m_worker.join ();
			}
		}
	}

private:
	TUniquePtr<TPromise<void>> m_promise;
	TFuture<void> m_future;
	std::thread m_worker;
};

bool bReadingGloveData = false;
ThreadController tcReadingGloveData;

bool bReadingOpticalData = false;
ThreadController tcReadingOpticalData;

bool bMonitorGlovePowerAndMagState = false;
ThreadController tcGlovePowerAndMagStateMonitor;

/**
 * Store Glove related information.
 *
 * 校准相关参数字段，需要在发生改变的时候及时更新。
 * 目前发生改变的几个时机：
 * 1、初始化
 * 2、发起B-Pose校准时
 * 3、B-Pose之后
 * 4、Reload校准文件/绑定关系文件
 */
struct Glove_Status
{
	EGlovePowerLevel power_level;
	int magneticed;	// 0-unmag | 1-maged | 2-unknown
	uint32 maged_time;	// 连续磁化累计时间
	EGloveMagneticedState mag_state;
	FString binded_dev_sn;
	OpticalSensorType sensor_type;
	FVector offset_to_sensor;
} gGloveStatus[GloveMod::GM_MaxNumb];

uint32 gLastViveDeviceCount = 0;
uint32 gValidViveDeviceCount = 0;

int goffsetindex = -1;
int gLeftHandTimes = 0;
int gRightHandTimes = 0;

CalibrationPercent CalibrationProgress;
GloveBVHData LeftHandData;
GloveBVHData RightHandData;

CalibrationPose CalibratingPos = GCP_Unknown;
int CalibrateTime = 0;
std::mutex gInternalMutex;

void Calibrationglove(CalibrationPose pos, int wait)
{
	CalibrationProgress.percent = 0;
	CalibratingPos = pos;
	CalibrateTime = wait;
	CalibrateThread = new std::thread(tCalibrateThread/*, NULL*/);
}

void tCalibrateThread()
{
	//bCalirating = true;
	if (hi5_isInViveSystem ())
	{
		CaliInVive (CalibratingPos, CalibrateTime);
	}
	else if (hi5_isInAliceSystem())
	{
		CaliInAlice (CalibratingPos, CalibrateTime);
	}
	else
	{
	}

	//bCalirating = false;
	hi5_cali_mutex_unlock ();
}

void CalibrationPrepare()
{
	if (hi5_isInViveSystem ())
	{
		calibrationPrepare ();
	}
	else if (hi5_isInAliceSystem())
	{
#if USE_ALICE
		if (IHybridData::GetPtr ())
		{
			AliceIns ().SendGloveCommand (CmdGlove::CG_CalibPrepare);
		}
#endif // USE_ALICE
	}
}

void CaliInVive(CalibrationPose pos, int wait/*ms*/)
{
	Glove_Status cached_left_state  = gGloveStatus[GM_LeftGlove];
	Glove_Status cached_right_state = gGloveStatus[GM_RightGlove];

	auto prolog = [&]() {
		if (pos == GCP_BPose)
		{
			//if (hi5_isLeftGloveAvailable ())
			{
				gGloveStatus[GM_LeftGlove].binded_dev_sn.Empty ();
				//gGloveStatus[GM_LeftGlove].offset_to_sensor.Set (0.0f, 0.0f, 0.0f);
				gGloveStatus[GM_LeftGlove].sensor_type = OST_Unknown;
			}
			//if (hi5_isRightGloveAvailable ())
			{
				gGloveStatus[GM_RightGlove].binded_dev_sn.Empty ();
				//gGloveStatus[GM_RightGlove].offset_to_sensor.Set (0.0f, 0.0f, 0.0f);
				gGloveStatus[GM_RightGlove].sensor_type = OST_Unknown;
			}
			UMotionCaptureFunctionLibrary::ResetDeviceId ();
		}
	};
	auto epilog = [&]() {
		if (pos == GCP_BPose)
		{
			FVector lOffset, rOffset;
			__hi5_getHandOffsetToOpticalDeviceInUE (lOffset, rOffset);

			bool leftOn  = false;
			bool rightOn = false;

			if (hi5_isLeftGloveAvailable ())
			{
				OpticalSensorType sensor_type;
				gGloveStatus[GM_LeftGlove].binded_dev_sn    = __hi5_getTrackerBindState (true, sensor_type);
				gGloveStatus[GM_LeftGlove].offset_to_sensor = lOffset;
				gGloveStatus[GM_LeftGlove].sensor_type      = sensor_type;
				leftOn                                      = true;
			}
			if (hi5_isRightGloveAvailable ())
			{
				OpticalSensorType sensor_type;
				gGloveStatus[GM_RightGlove].binded_dev_sn    = __hi5_getTrackerBindState (false, sensor_type);
				gGloveStatus[GM_RightGlove].offset_to_sensor = rOffset;
				gGloveStatus[GM_RightGlove].sensor_type      = sensor_type;
				rightOn                                      = true;
			}

			if (!leftOn && rightOn)
			{
				if ((cached_left_state.binded_dev_sn != gGloveStatus[GM_RightGlove].binded_dev_sn) &&
					(!cached_left_state.binded_dev_sn.IsEmpty ()))
				{
					setOptiSensorBindState (GM_LeftGlove, TCHAR_TO_ANSI (*cached_left_state.binded_dev_sn), cached_left_state.sensor_type);
					gGloveStatus[GM_LeftGlove].binded_dev_sn    = cached_left_state.binded_dev_sn;
					gGloveStatus[GM_LeftGlove].offset_to_sensor = cached_left_state.offset_to_sensor;
					gGloveStatus[GM_LeftGlove].sensor_type      = cached_left_state.sensor_type;
				}
			}
			if (leftOn && !rightOn)
			{
				if ((cached_right_state.binded_dev_sn != gGloveStatus[GM_LeftGlove].binded_dev_sn) &&
					(!cached_right_state.binded_dev_sn.IsEmpty()))
				{
					setOptiSensorBindState (GM_RightGlove, TCHAR_TO_ANSI (*cached_right_state.binded_dev_sn), cached_right_state.sensor_type);
					gGloveStatus[GM_RightGlove].binded_dev_sn   = cached_right_state.binded_dev_sn;
					gGloveStatus[GM_RightGlove].offset_to_sensor= cached_right_state.offset_to_sensor;
					gGloveStatus[GM_RightGlove].sensor_type     = cached_right_state.sensor_type;
				}
			}
			if (!leftOn && !rightOn)
			{
				if (!cached_left_state.binded_dev_sn.IsEmpty ())
				{
					setOptiSensorBindState (GM_LeftGlove, TCHAR_TO_ANSI (*cached_left_state.binded_dev_sn), cached_left_state.sensor_type);
					gGloveStatus[GM_LeftGlove].binded_dev_sn    = cached_left_state.binded_dev_sn;
					gGloveStatus[GM_LeftGlove].offset_to_sensor = cached_left_state.offset_to_sensor;
					gGloveStatus[GM_LeftGlove].sensor_type      = cached_left_state.sensor_type;
				}

				if (!cached_right_state.binded_dev_sn.IsEmpty ())
				{
					setOptiSensorBindState (GM_RightGlove, TCHAR_TO_ANSI (*cached_right_state.binded_dev_sn), cached_right_state.sensor_type);
					gGloveStatus[GM_RightGlove].binded_dev_sn    = cached_right_state.binded_dev_sn;
					gGloveStatus[GM_RightGlove].offset_to_sensor = cached_right_state.offset_to_sensor;
					gGloveStatus[GM_RightGlove].sensor_type      = cached_right_state.sensor_type;
				}
			}

			UMotionCaptureFunctionLibrary::CacheDeviceId ();
		}
	};

	prolog ();

	// calibration
	startCalibration(pos);

	const int sleepms = 50;
	const std::chrono::milliseconds dura (sleepms);

	// get percent
	int nPercent = 0;
	int waitCount = 0;
	int waitTotal = wait / sleepms + 1;
	while (nPercent != 100 && waitCount < waitTotal)
	{
		auto objPercent = getCalibratingPercent(pos);
		if (objPercent.percent != nPercent)
		{
			nPercent = objPercent.percent;
			if (nPercent == 100)
			{
				epilog ();
			}
		}
		CalibrationProgress = objPercent;

		if (nPercent == 100)
		{
			std::cout << "Done" << std::endl;
			break;
		}
		std::this_thread::sleep_for (dura);
		
		waitCount++;
	}
	if (nPercent != 100)
	{
		std::cout << "Time out" << std::endl;
	}
}

void CaliInAlice(CalibrationPose pose, int wait)
{
	// calibration
#if USE_ALICE
	if (IHybridData::GetPtr ()) {
		CmdGlove cglove;
		switch (pose) {
		case GCP_BPose:
			cglove = CmdGlove::CG_CalibBPose;
			break;
		case GCP_PPose:
			cglove = CmdGlove::CG_CalibPPose;
			break;
		default:
			// only support B-pose and P-pose
			return;
		}

		const std::chrono::seconds d (2);
		std::this_thread::sleep_for (d);

		if (!AliceIns ().SendGloveCommand (cglove))
		{
			return;
		}
	}
#else
	return;
#endif // USE_ALICE

	const int sleepms = 50;
	const std::chrono::milliseconds dura (sleepms);

	// get percent
	int nPercent = 0;
	int waitCount = 0;
	int waitTotal = wait / sleepms + 1;
	while (nPercent != 100 && waitCount < waitTotal)
	{
		int32 nCurrPercent;
		ECalibrationPose eCurrPose;
		IMotionCapture::Get ().GetCachedHybridDataCaliProgress (eCurrPose, nCurrPercent);
		if (nCurrPercent != nPercent)
		{
			nPercent = nCurrPercent;
			if (nPercent == 100)
			{
				//epilog ();
			}
		}
		CalibrationProgress.percent = nPercent;

		if (nPercent == 100)
		{
			std::cout << "Done" << std::endl;
			break;
		}
		std::this_thread::sleep_for (dura);

		waitCount++;
	}
	if (nPercent != 100)
	{
		std::cout << "Time out" << std::endl;
	}
}

int GetCalibrationProgressInPercent()
{
	//rg
	return FHi5_Manager_Thread::getInstance()->GetCalibrationpercent();
	//return Calibrationglove()
	//return CalibrationProgress.percent;
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

vr::IVRSystem* vrSystem () {
	
#if STEAMVR_MODULE_LOADED 
	if (ISteamVRPlugin::IsAvailable ())
	{
		static bool bSetPFn = false;
		if (!bSetPFn)
		{
			vr::VR_SetGenericInterfaceCallback (vr::VR_GetGenericInterface);
			bSetPFn = !bSetPFn;
		}

		//vrSys = ISteamVRPlugin::Get ().GetVRSystem (); // 💥 crash sometimes
		return vr::VRSystem ();
	}
#endif // STEAMVR_MODULE_LOADED 

	return nullptr;
}

const unsigned int max_tracker_number = vr::k_unMaxTrackedDeviceCount;

vr::TrackedDevicePose_t allTrackedDevice[max_tracker_number];

void initDevicePoses ()
{
	for (int i = 0; i < max_tracker_number; i++)
	{
		vr::TrackedDevicePose_t& pose = allTrackedDevice[i];

		pose.eTrackingResult = vr::TrackingResult_Uninitialized;
		pose.bPoseIsValid = false;
		pose.bDeviceIsConnected = false;
	}
}

void readOneFrameDevicePosesToBuffer (vr::TrackedDevicePose_t* TrackedDevices, size_t DeviceNumber = max_tracker_number)
{
	vr::IVRSystem* pSys = vrSystem ();
	if (pSys)
	{
		//pSys->GetDeviceToAbsoluteTrackingPose (vr::TrackingUniverseStanding, 0.0f, TrackedDevices, DeviceNumber);
		vr::IVRCompositor* pComp = vr::VRCompositor ();
		if (pComp)
		{
			//pComp->WaitGetPoses (TrackedDevices, DeviceNumber, NULL, 0);
			pComp->GetLastPoses (TrackedDevices, DeviceNumber, NULL, 0);
		}
	}
}

void readOneFrameDevicePoses ()
{
	readOneFrameDevicePosesToBuffer (allTrackedDevice);
}

bool hi5_isDeviceMissing (int id)
{
	if (id < 0 || id >= max_tracker_number)
	{
		return true;
	}

	vr::TrackedDevicePose_t& pose = allTrackedDevice[id];
	if (pose.bDeviceIsConnected && (pose.eTrackingResult == vr::ETrackingResult::TrackingResult_Running_OK))
	{
		return false;
	}
	else
	{
		return true;
	}
}

bool hi5_getDeviceType(int ID, ESteamVRTrackedDeviceType& Type)
{
	vr::IVRSystem* pSys = vrSystem ();
	if (pSys)
	{
		vr::ETrackedDeviceClass devClass = pSys->GetTrackedDeviceClass (ID);
		switch (devClass) {
		case vr::TrackedDeviceClass_Invalid:
			Type = ESteamVRTrackedDeviceType::Invalid;
			return true;

		case vr::TrackedDeviceClass_Controller:
			Type = ESteamVRTrackedDeviceType::Controller;
			return true;

		case vr::TrackedDeviceClass_TrackingReference:
			Type = ESteamVRTrackedDeviceType::TrackingReference;
			return true;

#if (ENGINE_MAJOR_VERSION >= 4)
#	if (ENGINE_MINOR_VERSION >= 16)
		case vr::TrackedDeviceClass_GenericTracker:
#	endif
#endif
		default:
			Type = ESteamVRTrackedDeviceType::Other;	// Vive Tracker is in this category
			return true;
		}
	}
	return false;
}




void hi5_tReadOpticalDataThreadBody()
{
	bReadingOpticalData = true;
	EGlovePositionSource pos_src = UMotionCaptureFunctionLibrary::GetGlovePositionSource ();

	std::chrono::milliseconds dura (5);

	if (pos_src == EGlovePositionSource::VivePosition)
	{
		initDevicePoses ();

		float pos[10], rot[10];
		FVector Position;
		FRotator Rotation;

		//while (bReadingOpticalData)
		{
			readOneFrameDevicePoses ();

			gValidViveDeviceCount = 0;

			for (int i = 1; i < max_tracker_number; i++)
			{
				vr::TrackedDevicePose_t& pose = allTrackedDevice[i];

				if (pose.bDeviceIsConnected && pose.eTrackingResult == vr::TrackingResult_Running_OK)
				{
					OpticalSensorType sensor_type;
					ESteamVRTrackedDeviceType dev_type;
					if (hi5_getDeviceType (i, dev_type))
					{
						if (dev_type == ESteamVRTrackedDeviceType::Controller)
						{
							sensor_type = OST_HTC_VIVE_Controller;
						}
						else if (dev_type == ESteamVRTrackedDeviceType::Other)
						{
							sensor_type = OST_HTC_VIVE_Tracker;
						}
						else {
							continue;
						}

						FString sn;
						if (hi5_getViveTrackerSN (i, sn))
						{
							if (!bCalirating && UMotionCaptureFunctionLibrary::GetTrackedDeviceDataInViveSpace (i, pos, rot))
							{
								pushOpticalSensorData (TCHAR_TO_ANSI (*sn), sensor_type, pos, rot);
							}
#if STEAMVR_MODULE_LOADED
							else if (bCalirating && USteamVRFunctionLibrary::GetTrackedDevicePositionAndOrientation (i, Position, Rotation))
							{
								UMotionCaptureFunctionLibrary::UESpace2ViveSpace (Position, Rotation, pos, rot);
								pushOpticalSensorData (TCHAR_TO_ANSI (*sn), sensor_type, pos, rot);
							}
#endif
						}
					}
				}

				if (pose.bDeviceIsConnected)
				{
					gValidViveDeviceCount++;
				}
			}

			if (gValidViveDeviceCount != gLastViveDeviceCount)
			{
				UMotionCaptureFunctionLibrary::CacheDeviceId ();
				gLastViveDeviceCount = gValidViveDeviceCount;
			}

			//Sleep (5);	// 180fps
			std::this_thread::sleep_for (dura);
		}
	}
}

//////////////////////////////////////////////

FMatrix ToFMatrix (const vr::HmdMatrix34_t& tm)
{
	// Rows and columns are swapped between vr::HmdMatrix34_t and FMatrix
	return FMatrix (
		FPlane (tm.m[0][0], tm.m[1][0], tm.m[2][0], 0.0f),
		FPlane (tm.m[0][1], tm.m[1][1], tm.m[2][1], 0.0f),
		FPlane (tm.m[0][2], tm.m[1][2], tm.m[2][2], 0.0f),
		FPlane (tm.m[0][3], tm.m[1][3], tm.m[2][3], 1.0f));
}

bool hi5_getTrackedDevicePose(int unTrackedDevice, FVector& Position, FRotator& Rotator)
{
	//rg ？？ 
	if (hi5_isDeviceMissing (unTrackedDevice))
	{
		return false;
	}
	//rg 没往下走
	vr::TrackedDevicePose_t& pose = allTrackedDevice[unTrackedDevice];
	{	// PoseToOrientation
		FMatrix Pose = ToFMatrix (pose.mDeviceToAbsoluteTracking);
		FQuat Orientation (Pose);

		FQuat OutOrientation;
		OutOrientation.X = -Orientation.Z;
		OutOrientation.Y = Orientation.X;
		OutOrientation.Z = Orientation.Y;
		OutOrientation.W = -Orientation.W;

		Position = ((FVector (-Pose.M[3][2], Pose.M[3][0], Pose.M[3][1])) * 100);
		OutOrientation.Normalize ();
		Rotator = OutOrientation.Rotator ();
	}
	return true;
}


////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

void tDeviceReadThead(/*void* param*/)
{
	//enableFingerAdbFixed(true);
	std::chrono::milliseconds dura (5);

	GloveBVHData bvhData; // left and right glove data

	int Ret = 0;
	bReadingGloveData = true;
	while (bReadingGloveData)
	{
		// read status
		Ret = isDongleAvailable();
		if (Ret == 0)
		{
			std::this_thread::sleep_for (std::chrono::milliseconds (10));
			continue;
		}

		readBVHData (&bvhData);
		if (!bvhData.bUpdate)
		{
			std::this_thread::sleep_for (dura);
			continue;
		}

		NetworkManager::GetInstance()->OnLocalPlayerBVHDataReceived((uint8*)&bvhData, sizeof(bvhData));
	}
}

void hi5_tGlovePowerAndMagStateMonitorBody ()
{
	const uint32 monitor_interval = 100;
	const std::chrono::milliseconds dura (monitor_interval);

	/// monitor_func implements a simple state machine.
	auto monitor_func = [monitor_interval](GloveMod Mod)
	{
		Glove_Status& gs = gGloveStatus[Mod];

		if (gs.magneticed == 0)
		{
			gs.mag_state = EGloveMagneticedState::Normal;
			gs.maged_time = 0;
		}
		else
		{
			if (gs.mag_state == EGloveMagneticedState::Warn)
			{
				gs.maged_time++;
				if ((gs.maged_time * monitor_interval) >= 15000)
				{
					gs.mag_state = EGloveMagneticedState::Bad;
				}
			}
			else if (gs.mag_state != EGloveMagneticedState::Bad)
			{
				gs.mag_state = EGloveMagneticedState::Warn;
			}
		}
	};

	bool isLeftLastOn             = hi5_isLeftGloveAvailable ();
	bool isRightLastOn            = hi5_isRightGloveAvailable ();

	bMonitorGlovePowerAndMagState = true;
	while (bMonitorGlovePowerAndMagState)
	{
		if (hi5_isLeftGloveAvailable ())
		{
			monitor_func (GM_LeftGlove);
			isLeftLastOn = true;
		}
		if (!hi5_isLeftGloveAvailable () && isLeftLastOn)
		{
			isLeftLastOn = false;
			gGloveStatus[GM_LeftGlove].magneticed = 0;
			gGloveStatus[GM_LeftGlove].maged_time = 0;
			gGloveStatus[GM_LeftGlove].mag_state  = EGloveMagneticedState::Unknown;

			gGloveStatus[GM_LeftGlove].power_level = EGlovePowerLevel::Unknown;
		}

		if (hi5_isRightGloveAvailable ())
		{
			monitor_func (GM_RightGlove);
			isRightLastOn = true;
		}
		if (!hi5_isRightGloveAvailable () && isRightLastOn)
		{
			isRightLastOn = false;
			gGloveStatus[GM_RightGlove].magneticed = 0;
			gGloveStatus[GM_RightGlove].maged_time = 0;
			gGloveStatus[GM_RightGlove].mag_state  = EGloveMagneticedState::Unknown;

			gGloveStatus[GM_RightGlove].power_level = EGlovePowerLevel::Unknown;
		}

		std::this_thread::sleep_for (dura);
	}
}

int StartReadGlove()
{
	FHi5_Manager_Thread::getInstance()->SetEnableFingerAdbFixed(false);
	FHi5_Manager_Thread::getInstance()->RequestConnect(false);
	hi5_initGloveStatus ();
	
	//rg
	// FGameDelegates::Get ().GetExitCommandDelegate ().AddLambda ([]() {__hi5_stopThreads (); });
	//
	
	// setDongleStateChangedHandle (nullptr, hi5_fnDongleStateChanged);
	// setGloveStateChangedHandle (nullptr, hi5_fnGloveStateChanged);
	// tcGlovePowerAndMagStateMonitor.run (hi5_tGlovePowerAndMagStateMonitorBody);
	//
	// hi5DongleStart ();
	//tcReadingGloveData.run (tDeviceReadThead);
	//tcReadingOpticalData.run (hi5_tReadOpticalDataThreadBody);
	
	return 0;
}

void StopReadGlove()
{
	FHi5_Manager_Thread::getInstance()->RequestCloseConnect();
	return;
	__hi5_stopThreads ();

	hi5DongleStop ();
	//UHybridDataHelperFunctionLibrary::ClearDataReceivers();
}

void ModifyDataRotate(GloveBVHData& bvhData, int offset, float valueoffset)
{
	int DataSize;
	int RotateOffset;
	int OffsetOffset = 0;
	if (bvhData.bWithDisp)
	{
		DataSize = 6;
		RotateOffset = 3;
	}
	else
	{
		DataSize = 3;
		RotateOffset = 3;
	}
	int DataCount = bvhData.dataCount / DataSize;
	if (!bvhData.bWithDisp)
	{
		DataCount--;
	}
	float *RotationOffset = &bvhData.data[RotateOffset];
	for (int i = 0; i < DataCount; i++)
	{
		RotationOffset[i*DataSize + offset] += valueoffset;
	}
}

int GetGloveData(GloveMod LeftOrRight, uint8 *data)
{
	if (LeftOrRight == GM_RightGlove)
	{
		memcpy(data,&RightHandData, sizeof(RightHandData));
		return sizeof(RightHandData);
	}
	else if (LeftOrRight == GM_LeftGlove)
	{
		memcpy(data,&LeftHandData, sizeof(LeftHandData));
		return sizeof(LeftHandData);
	}
	return 0;
}

bool WCharToChar(wchar_t* strWChar,int WCharLen, char *OutValue, int& outsize)
{
	int len = WCharLen;
	outsize = len * 2 + 2;
	char *pTemp = OutValue;
	memset(pTemp, 0, outsize);
//#ifdef ANDROID
	int i = 0;
	for (i = 0; i < len; i++)
	{
		pTemp[i] = strWChar[i];
	}
	pTemp[len] = 0;
	outsize = len;
//#else
//	len = ::WideCharToMultiByte(CP_ACP, 0, strWChar, len, pTemp, len, NULL, NULL);
//	outsize = len;
//#endif
	return true;
}

void pushGloveTrackerData(bool isLeft, float* pos/*xyz*/, float* quat/*sxyz*/)
{
	//pushTrackerData(isLeft, pos, quat);
}

FTransform GetChildWorldTransform(const FTransform& ParentWorldTransform, const FTransform& RelativeTransform)
{
	FTransform ChildWorldTransform;
	FVector RelativeLocation = RelativeTransform.GetLocation();
	FQuat ParentQuat = ParentWorldTransform.GetRotation();
	FVector Offset = ParentQuat.RotateVector(RelativeLocation);
	FVector ChildLocation = ParentWorldTransform.GetLocation() + Offset;
	ChildWorldTransform.SetLocation(ChildLocation);
	ChildWorldTransform.SetRotation(RelativeTransform.GetRotation()*ParentWorldTransform.GetRotation());
	ChildWorldTransform.SetScale3D(ParentWorldTransform.GetScale3D()*RelativeTransform.GetScale3D());
	return ChildWorldTransform;
}

void RegistViveTrackCallBack(bool IsLeftHand, std::function<bool(float*, float*) > func)
{
	if (IsLeftHand)
	{
		LeftCallback = func;
	}
	else
	{
		RightCallback = func;
	}
}

void hi5_enableVibratorOfLeft (unsigned int leftGloveTimeSpan/*ms*/)
{
	enableLeftGloveVibrator (leftGloveTimeSpan);
}

void hi5_enableVibratorOfRight (unsigned int rightGloveTimeSpan/*ms*/)
{
	enableRightGloveVibrator (rightGloveTimeSpan);
}

void hi5_enableVibratorOfBoth (unsigned int leftGloveTimeSpan, unsigned int rightGloveTimeSpan)
{
	enableGlovesVibrator (leftGloveTimeSpan, rightGloveTimeSpan);
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

bool hi5_getViveTrackerSN(const int32 index, FString& sn)
{
	sn.Reset ();

	if (index < 0) {
		return false;
	}

	vr::IVRSystem* pSys = vrSystem ();
	if (pSys)
	{
		vr::TrackedDeviceIndex_t unDeviceIndex = index;
		vr::ETrackedDeviceProperty prop = vr::Prop_SerialNumber_String;
		char pchValue[/*vr::k_unTrackingStringSize*/32];
		uint32_t ret = pSys->GetStringTrackedDeviceProperty (unDeviceIndex, prop, pchValue, /*vr::k_unTrackingStringSize*/32);
		if (ret == 0) {
			return false;
		}
		else {
			sn.AppendChars (ANSI_TO_TCHAR (pchValue), ret);
			return true;
		}
	}
	else
	{
		return false;
	}
}

bool hi5_getViveTrackerId (const FString& sn, int32& id)
{
	id = -1;
	if (sn.IsEmpty())
	{
		return false;
	}
	FString currSN;
	for (unsigned int i = 0; i < max_tracker_number; i++)
	{
		if (hi5_getViveTrackerSN (i, currSN) && sn == currSN)
		{
			id = i;
			return true;
		}
	}
	return false;
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

void __stdcall hi5_fnGloveStateChanged (void* userData, GloveMod glove, GloveEvent event)
{
	Glove_Status& gs = gGloveStatus[glove];

	switch (event)
	{
	//case GE_PowerKeyClicked: // Power键的单击事件（长按、双击等都不会触发）
	//	break;
	case GE_LowPower:       // Low power
		gs.power_level = EGlovePowerLevel::Low;
		break;
	case GE_NormalPower:    // Normal power
		gs.power_level = EGlovePowerLevel::Normal;
		break;
	case GE_FullPower:      // Full power
		gs.power_level = EGlovePowerLevel::Full;
		break;
	case GE_NotMagneticed:  // 未被磁化
		gs.magneticed = 0;
		break;
	case GE_Magneticed:     // 已被磁化
		gs.magneticed = 1;
		break;
	case GE_Unknown:
	default:
		break;
	}
	//std::cout << "### " << strGlove << strAct << std::endl;
}

void __stdcall hi5_fnDongleStateChanged (void *userData, const DongleIdentity *hid, DongleEvent action) {
	switch (action)
	{
	case DE_Insert:
		UMotionCaptureFunctionLibrary::LoadCalibrationData ();
		break;

	case DE_Removed:
		break;
	}

	const char *strAct = (action == DE_Insert ? "inserted" : "removed");
	std::cout << "### Dongle(SN:" << hid->SerialNumber << ") " << strAct
		<< std::endl;
}

void hi5_initGloveStatus()
{
	hi5_cali_mutex_lock ();
	for (int i = 0; i < sizeof gGloveStatus / sizeof Glove_Status; i++)
	{
		gGloveStatus[i].power_level      = EGlovePowerLevel::Unknown;
		gGloveStatus[i].magneticed       = 0;
		gGloveStatus[i].maged_time       = 0;
		gGloveStatus[i].mag_state        = EGloveMagneticedState::Unknown;
		gGloveStatus[i].sensor_type      = OpticalSensorType::OST_Unknown;
		gGloveStatus[i].offset_to_sensor = FVector (0.0f, 0.0f, 0.0f);
	}
	hi5_cali_mutex_unlock ();
}

FString __hi5_getTrackerBindState (bool left, OpticalSensorType& sensor_type)
{
	GloveMod mod = left ? GloveMod::GM_LeftGlove : GloveMod::GM_RightGlove;
	char sn[64] = { 0 };
	if (getOptiSensorBindState (mod, sn, 64, &sensor_type) != 0)
	{
		return FString (ANSI_TO_TCHAR (sn));
	}
	else {
		return FString();
	}
}

FString hi5_getTrackerBindState(bool left, OpticalSensorType& sensor_type)
{
	GloveMod mod = left ? GloveMod::GM_LeftGlove : GloveMod::GM_RightGlove;
	sensor_type = gGloveStatus[mod].sensor_type;
	return gGloveStatus[mod].binded_dev_sn;
}

void hi5_loadTrackerBindState (bool left, const FString& sn, const OpticalSensorType& sensor_type)
{
	hi5_cali_mutex_lock ();

	GloveMod mod = left ? GloveMod::GM_LeftGlove : GloveMod::GM_RightGlove;
	setOptiSensorBindState (mod, TCHAR_TO_ANSI (*sn), sensor_type);

	// epilog
	gGloveStatus[mod].binded_dev_sn = sn;
	gGloveStatus[mod].sensor_type   = sensor_type;

	UMotionCaptureFunctionLibrary::CacheDeviceId ();

	hi5_cali_mutex_unlock ();
}

bool hi5_loadCalibrationFile(char* filePath)
{
	hi5_cali_mutex_lock ();

	bool ret = loadCalibrationData (filePath) != 0;
	//__hi5_getHandOffsetToOpticalDeviceInUE (gGloveStatus[GM_LeftGlove].offset_to_sensor, gGloveStatus[GM_RightGlove].offset_to_sensor);
	//patch_wrist_offset ();

	hi5_cali_mutex_unlock ();
	return ret;
}

bool hi5_isDongleAvailable()
{
	return isDongleAvailable () == TRUE;
}

bool hi5_isLeftGloveAvailable()
{
	HI5GloveStatus status;
	readDeviceStatus (&status);
	bool ret = (status == DV_LeftGloveAvailable) || (status == DV_BothGloveAvailable);
	return ret;
}

bool hi5_isRightGloveAvailable()
{
	HI5GloveStatus status;
	readDeviceStatus (&status);
	bool ret = (status == DV_RightGloveAvailable) || (status == DV_BothGloveAvailable);
	return ret;
}

void __hi5_stopThreads()
{
	bReadingOpticalData = false;
	tcReadingOpticalData.waitForFinished ();

	bReadingGloveData = false;
	tcReadingGloveData.waitForFinished ();

	bMonitorGlovePowerAndMagState = false;
	tcGlovePowerAndMagStateMonitor.waitForFinished ();
}

void __hi5_getHandOffsetToOpticalDeviceInUE(FVector& LeftPosition, FVector& RightPosition)
{
	float posL[3], posR[3];
	posL[0] = FHi5_Manager_Thread::leftTrackerInitOffset[0];
	posL[1] = FHi5_Manager_Thread::leftTrackerInitOffset[1];
	posL[2] = FHi5_Manager_Thread::leftTrackerInitOffset[2];
	posR[0] = FHi5_Manager_Thread::rightTrackerInitOffset[0];
	posR[1] = FHi5_Manager_Thread::rightTrackerInitOffset[1];
	posR[2] = FHi5_Manager_Thread::rightTrackerInitOffset[2];
	//getWristPositionInTrackerFrame (posL, posR);

	//FMotionCaptureHelper::GetVivePosition(posL, 0)

	LeftPosition.Y = posL[0] * 100;
	LeftPosition.Z = posL[1] * 100;
	LeftPosition.X = -posL[2] * 100;

	RightPosition.Y = posR[0] * 100;
	RightPosition.Z = posR[1] * 100;
	RightPosition.X = -posR[2] * 100;
}

void hi5_getHandOffsetToOpticalDeviceInUE(FVector& LeftPosition, FVector& RightPosition)
{
	LeftPosition = gGloveStatus[GM_LeftGlove].offset_to_sensor;
	RightPosition = gGloveStatus[GM_RightGlove].offset_to_sensor;
}

void hi5_cali_mutex_lock()
{
	//gInternalMutex.lock ();
}

void hi5_cali_mutex_unlock()
{
	//gInternalMutex.unlock ();
}

bool hi5_isInViveSystem()
{
	return (UMotionCaptureFunctionLibrary::GetGlovePositionSource () == EGlovePositionSource::VivePosition);
}

bool hi5_isInAliceSystem()
{
	return (UMotionCaptureFunctionLibrary::GetGlovePositionSource () == EGlovePositionSource::AlicePosition);
}

void hi5_getGlovePowerLevel (EGlovePowerLevel& left, EGlovePowerLevel& right)
{
	left = gGloveStatus[GloveMod::GM_LeftGlove].power_level;
	right = gGloveStatus[GloveMod::GM_RightGlove].power_level;
}

void hi5_getGloveMag(EGloveMagneticedState& left, EGloveMagneticedState& right)
{
	left  = gGloveStatus[GloveMod::GM_LeftGlove].mag_state;
	right = gGloveStatus[GloveMod::GM_RightGlove].mag_state;
}

// for test
void patch_wrist_offset ()
{
	float posL[3], posR[3];
	getWristPositionInTrackerFrame (posL, posR);

	if (!(
		(posL[0] > -0.000001 && posL[0] < 0.000001) &&
		(posL[1] > -0.000001 && posL[1] < 0.000001) &&
		(posL[2] > -0.000001 && posL[2] < 0.000001)
		))
	{
		FVector& LeftPosition = gGloveStatus[GM_LeftGlove].offset_to_sensor;
		LeftPosition.Y = posL[0] * 100;
		LeftPosition.Z = posL[1] * 100;
		LeftPosition.X = -posL[2] * 100;
	}

	if (!(
		(posR[0] > -0.000001 && posR[0] < 0.000001) &&
		(posR[1] > -0.000001 && posR[1] < 0.000001) &&
		(posR[2] > -0.000001 && posR[2] < 0.000001)
		))
	{
		FVector& RightPosition = gGloveStatus[GM_RightGlove].offset_to_sensor;
		RightPosition.Y = posR[0] * 100;
		RightPosition.Z = posR[1] * 100;
		RightPosition.X = -posR[2] * 100;
	}
}

void JudgeLeftHandAndRightHDeviceId(APawn* inPawn)
{
	initDevicePoses();
	readOneFrameDevicePoses ();
	
	//TArray 
	for (int i = 1; i < max_tracker_number; i++)
	{
		vr::TrackedDevicePose_t& pose = allTrackedDevice[i];
		if (pose.bDeviceIsConnected && pose.eTrackingResult == vr::TrackingResult_Running_OK)
		{
			OpticalSensorType sensor_type;
			ESteamVRTrackedDeviceType dev_type;
			if (hi5_getDeviceType (i, dev_type))
			{
				// if (dev_type == ESteamVRTrackedDeviceType::Controller)
				// {
				// 	sensor_type = OST_HTC_VIVE_Controller;
				// }
				// else
				if (dev_type == ESteamVRTrackedDeviceType::Other)
				{
					sensor_type = OST_HTC_VIVE_Tracker;
					FString sn;
					if (hi5_getViveTrackerSN (i, sn))
					{
						//float pos[10], rot[10];
						FVector Position;
						FRotator Rotation;
						if (hi5_getTrackedDevicePose (i, Position, Rotation))
						//if(UMotionCaptureFunctionLibrary::GetTrackedDeviceDataInViveSpace (i, pos, rot))
						{
							if(FHi5_Manager_Thread::getInstance() != nullptr)
							{
								// Position[0] = pos[0];
								// Position[1] = pos[1];
								// Position[2] = pos[2];
								// Rotation.W = rot[0];
								// Rotation.X= rot[3];
								// Rotation.Y = rot[2];
								// Rotation.Z = rot[1];
								// rot[0] = q.W;	// vive_w <= ue_w
								// rot[1] = -q.Y;	// vive_x <= -ue_y
								// rot[2] = -q.Z;	// vive_y <= -ue_z
								// rot[3] = q.X;	// vive_z <= ue_x
								FVive_Tracker_Infor_Map&  Tracker_Infor_Ma = FHi5_Manager_Thread::getInstance()->GetVive_Tracker_Infor_Map();
								if(!Tracker_Infor_Ma.IsContain(i))
								{
									FVive_Tracker_Infor item;
									item.trackerId = i;
									item.trackerIdentity = sn;
									item.position = Position;
									item.rotation = Rotation;
									Tracker_Infor_Ma.AddItem(item);
								}
								else
								{
									Tracker_Infor_Ma.SetPosAndQua(i,Position,Rotation);
								}
							}
						}
					}
				}
				else
				{
					continue;
				}
			}
		}
	}

	//判断左右手trackeid
	if(FHi5_Manager_Thread::getInstance() != nullptr)
	{
		FHi5_Manager_Thread::getInstance()->GetVive_Tracker_Infor_Map().FindTwoDevices(inPawn);
	}
}



void hi5_2tReadOpticalDataBody()
{
	bReadingOpticalData = true;
	EGlovePositionSource pos_src = UMotionCaptureFunctionLibrary::GetGlovePositionSource ();

	//std::chrono::milliseconds dura (5);

	if (pos_src == EGlovePositionSource::VivePosition)
	{
		initDevicePoses ();

		float pos[10], rot[10];
		FVector Position;
		FRotator Rotation;

		//while (bReadingOpticalData)
		{
			readOneFrameDevicePoses ();

			gValidViveDeviceCount = 0;

			for (int i = 1; i < max_tracker_number; i++)
			{
				vr::TrackedDevicePose_t& pose = allTrackedDevice[i];

				if (pose.bDeviceIsConnected && pose.eTrackingResult == vr::TrackingResult_Running_OK)
				{
					OpticalSensorType sensor_type;
					ESteamVRTrackedDeviceType dev_type;
					if (hi5_getDeviceType (i, dev_type))
					{
						if (dev_type == ESteamVRTrackedDeviceType::Controller)
						{
							sensor_type = OST_HTC_VIVE_Controller;
						}
						else if (dev_type == ESteamVRTrackedDeviceType::Other)
						{
							sensor_type = OST_HTC_VIVE_Tracker;
						}
						else {
							continue;
						}

						FString sn;
						if (hi5_getViveTrackerSN (i, sn))
						{
							if (!bCalirating && UMotionCaptureFunctionLibrary::GetTrackedDeviceDataInViveSpace (i, pos, rot))
							{
								//pushOpticalSensorData (TCHAR_TO_ANSI (*sn), sensor_type, pos, rot);
								std::string sensorIdentity(TCHAR_TO_ANSI (*sn));
								float fpos[3];
								fpos[0] = pos[0];
								fpos[1] = pos[1];
								fpos[2] = pos[2];
								float frot[4];
								frot[0] = rot[0];
								frot[1] = rot[1];
								frot[2] = rot[2];
								frot[3] = rot[3];
								
								FHi5_Manager_Thread::getInstance()->AddOpticData( sensorIdentity, sensor_type, fpos, frot);
							}
#if STEAMVR_MODULE_LOADED
							else if (bCalirating && USteamVRFunctionLibrary::GetTrackedDevicePositionAndOrientation (i, Position, Rotation))
							{
								UMotionCaptureFunctionLibrary::UESpace2ViveSpace (Position, Rotation, pos, rot);
								//pushOpticalSensorData (TCHAR_TO_ANSI (*sn), sensor_type, pos, rot);
								std::string sensorIdentity(TCHAR_TO_ANSI (*sn));
								float fpos[3];
								fpos[0] = pos[0];
								fpos[1] = pos[1];
								fpos[2] = pos[2];
								float frot[4];
								 frot[0] = rot[0];
								 frot[1] = rot[1];
								 frot[2] = rot[2];
								frot[3] = rot[3];
								
								FHi5_Manager_Thread::getInstance()->AddOpticData( sensorIdentity, sensor_type, fpos, frot);
							}
#endif
						}
					}
				}

				if (pose.bDeviceIsConnected)
				{
					gValidViveDeviceCount++;
				}
			}

			if (gValidViveDeviceCount != gLastViveDeviceCount)
			{
				UMotionCaptureFunctionLibrary::CacheDeviceId ();
				gLastViveDeviceCount = gValidViveDeviceCount;
			}

			//Sleep (5);	// 180fps
			//std::this_thread::sleep_for (dura);
		}
	}
}

void hi5_2PreCalibration()
{
	//if (hi5_isLeftGloveAvailable ())
	{
		gGloveStatus[GM_LeftGlove].binded_dev_sn.Empty ();
		//gGloveStatus[GM_LeftGlove].offset_to_sensor.Set (0.0f, 0.0f, 0.0f);
		gGloveStatus[GM_LeftGlove].sensor_type = OST_Unknown;
	}
	//if (hi5_isRightGloveAvailable ())
	{
		gGloveStatus[GM_RightGlove].binded_dev_sn.Empty ();
		//gGloveStatus[GM_RightGlove].offset_to_sensor.Set (0.0f, 0.0f, 0.0f);
		gGloveStatus[GM_RightGlove].sensor_type = OST_Unknown;
	}
	UMotionCaptureFunctionLibrary::ResetDeviceId ();
}

void hi5_2CalibrationEpilog(FString snLeft, FString snRight)
{
	Glove_Status cached_left_state  = gGloveStatus[GM_LeftGlove];
	Glove_Status cached_right_state = gGloveStatus[GM_RightGlove];
	//if (pos == GCP_BPose)
		{
			FVector lOffset, rOffset;
			__hi5_getHandOffsetToOpticalDeviceInUE (lOffset, rOffset);

			bool leftOn  = false;
			bool rightOn = false;

			//if (hi5_isLeftGloveAvailable ())
			{
				//OpticalSensorType sensor_left_type;
				gGloveStatus[GM_LeftGlove].binded_dev_sn    = snLeft;
				gGloveStatus[GM_LeftGlove].offset_to_sensor = lOffset;
				gGloveStatus[GM_LeftGlove].sensor_type      = OpticalSensorType::OST_HTC_VIVE_Tracker;
				leftOn                                      = true;
			}
			//if (hi5_isRightGloveAvailable ())
			{
				//OpticalSensorType sensor_type;
				gGloveStatus[GM_RightGlove].binded_dev_sn    = snRight;
				gGloveStatus[GM_RightGlove].offset_to_sensor = rOffset;
				gGloveStatus[GM_RightGlove].sensor_type      = OpticalSensorType::OST_HTC_VIVE_Tracker;
				rightOn                                      = true;
			}

			/*if (!leftOn && rightOn)
			{
				if ((cached_left_state.binded_dev_sn != gGloveStatus[GM_RightGlove].binded_dev_sn) &&
					(!cached_left_state.binded_dev_sn.IsEmpty ()))
				{
					setOptiSensorBindState (GM_LeftGlove, TCHAR_TO_ANSI (*cached_left_state.binded_dev_sn), cached_left_state.sensor_type);
					gGloveStatus[GM_LeftGlove].binded_dev_sn    = cached_left_state.binded_dev_sn;
					gGloveStatus[GM_LeftGlove].offset_to_sensor = cached_left_state.offset_to_sensor;
					gGloveStatus[GM_LeftGlove].sensor_type      = cached_left_state.sensor_type;
				}
			}
			if (leftOn && !rightOn)
			{
				if ((cached_right_state.binded_dev_sn != gGloveStatus[GM_LeftGlove].binded_dev_sn) &&
					(!cached_right_state.binded_dev_sn.IsEmpty()))
				{
					setOptiSensorBindState (GM_RightGlove, TCHAR_TO_ANSI (*cached_right_state.binded_dev_sn), cached_right_state.sensor_type);
					gGloveStatus[GM_RightGlove].binded_dev_sn   = cached_right_state.binded_dev_sn;
					gGloveStatus[GM_RightGlove].offset_to_sensor= cached_right_state.offset_to_sensor;
					gGloveStatus[GM_RightGlove].sensor_type     = cached_right_state.sensor_type;
				}
			}
			if (!leftOn && !rightOn)
			{
				if (!cached_left_state.binded_dev_sn.IsEmpty ())
				{
					setOptiSensorBindState (GM_LeftGlove, TCHAR_TO_ANSI (*cached_left_state.binded_dev_sn), cached_left_state.sensor_type);
					gGloveStatus[GM_LeftGlove].binded_dev_sn    = cached_left_state.binded_dev_sn;
					gGloveStatus[GM_LeftGlove].offset_to_sensor = cached_left_state.offset_to_sensor;
					gGloveStatus[GM_LeftGlove].sensor_type      = cached_left_state.sensor_type;
				}

				if (!cached_right_state.binded_dev_sn.IsEmpty ())
				{
					setOptiSensorBindState (GM_RightGlove, TCHAR_TO_ANSI (*cached_right_state.binded_dev_sn), cached_right_state.sensor_type);
					gGloveStatus[GM_RightGlove].binded_dev_sn    = cached_right_state.binded_dev_sn;
					gGloveStatus[GM_RightGlove].offset_to_sensor = cached_right_state.offset_to_sensor;
					gGloveStatus[GM_RightGlove].sensor_type      = cached_right_state.sensor_type;
				}
			}*/
			
			UMotionCaptureFunctionLibrary::CacheDeviceId ();
		}
}
