// Copyright 2015 Noitom Technology Co., Ltd. All Rights Reserved.

#include "Hi5PacketParser.h"
#include "Runtime/Launch/Resources/Version.h"

namespace Hi5
{
	FPacketParser::FPacketParser( ) :
		FMotionCaptureHelper( ),
		CurMaxPacketID( 0 )
	{

	}

	FPacketParser::~FPacketParser( )
	{

	}

	bool FPacketParser::Read( const FArrayReaderPtr& Data, FAnimationDataPtr AnimData, int32& AvatarID, FName& AvatarName )
	{
		if (!CheckHeader( Data ))
		{
			return false;
		}

		AvatarID = PacketHeader.AvatarID;
		AvatarName = PacketHeader.AvatarName;

		const uint8* RawData = Data->GetData( );
		int Offset = FHi5PacketHeader::HeaderSize;

		if (0 != PacketHeader.WithReference)
		{
			FVector RootPosition = GetPosition( RawData, Offset );
			AnimData->RootPosition = FVector( RootPosition.X, RootPosition.Z, RootPosition.Y );
			AnimData->RootRotation = GetQuaternion( RawData, Offset );
			AnimData->bUseRootEffect = true;
		}

		FVector HipPos = GetPosition( RawData, Offset );
		AnimData->BonePositions[EMCBones::RightHand] = HipPos;
		FQuat HipQuat = GetQuaternion( RawData, Offset );
		AnimData->BoneRotations[EMCBones::RightHand] = HipQuat;

		if (0 != PacketHeader.WithDisplacement)
		{
			AnimData->bWithDisplacement = true;
		}
		else
		{
			AnimData->bWithDisplacement = false;
		}

		for (int i = 1; i < FPacketParser::PNBoneCount; i++)
		{
			ParseAnimData( RawData, i, Offset, AnimData );
		}

		return true;
	}

	bool FPacketParser::CheckHeader( const FArrayReaderPtr& Data )
	{
		if (FHi5PacketHeader::HeaderSize > Data->Num( ))
		{
			return false;
		}

		if (!ParseHeader( Data ))
		{
			PN_LOG( Error, TEXT( "Parse packet header failed." ) );
			return false;
		}

		if ((FHi5PacketHeader::ValidHeaderBegin != PacketHeader.HeaderBegin) ||
			(FHi5PacketHeader::ValidHeaderEnd != PacketHeader.HeaderEnd))
		{
			return false;
		}

		if (PacketHeader.Sequence > CurMaxPacketID)
		{
			CurMaxPacketID = PacketHeader.Sequence;
		}
		else
		{
			if (PacketHeader.Sequence < CurMaxPacketID / 2)	// loop
			{
				CurMaxPacketID = PacketHeader.Sequence;
			}
		}

		if (Data->Num( ) != (PacketHeader.DataCount * 4 + FHi5PacketHeader::HeaderSize))
		{
			return false;
		}

		return true;
	}

	void FPacketParser::ParseAnimData( const uint8* Data, const int32 Index, int32& Offset, FAnimationDataPtr AnimData )
	{
		uint8 PNBoneIndex = PNBoneIndices[Index];

		if (0 != PacketHeader.WithDisplacement)
		{
			FVector BonePos = GetPosition( Data, Offset );
			AnimData->BonePositions[PNBoneIndex] = BonePos;
		}
		FQuat BoneQuat = GetQuaternion( Data, Offset );
		AnimData->BoneRotations[PNBoneIndex] = BoneQuat;
	}

	bool FPacketParser::ParseHeader( const FArrayReaderPtr& Data )
	{
		const uint8* RawData = Data->GetData( );
		int32 Offset = 0;
		for (TFieldIterator<FProperty> PropIt( FHi5PacketHeader::StaticStruct( ), EFieldIteratorFlags::ExcludeSuper ); PropIt; ++PropIt)
		{
			FProperty* Property = *PropIt;

			if (Property->IsA( FByteProperty::StaticClass( ) ))
			{
				uint8 Value = GetUInt8( RawData, Offset );
				FByteProperty* ByteProp = CastField<FByteProperty>( Property );
				ByteProp->SetPropertyValue( (void*)(Property->ContainerPtrToValuePtr<uint8>( &PacketHeader )), Value );
			}
			else if (Property->IsA( FInt16Property::StaticClass( ) ))
			{
				int16 Value = GetUInt16( RawData, Offset );
				FInt16Property* Int16Prop = CastField<FInt16Property>( Property );
				Int16Prop->SetPropertyValue( (void*)(Property->ContainerPtrToValuePtr<int16>( &PacketHeader )), Value );
			}
			else if (Property->IsA( FUInt16Property::StaticClass( ) ))
			{
				uint16 Value = GetUInt16( RawData, Offset );
				FUInt16Property* UInt16Prop = CastField<FUInt16Property>( Property );
				UInt16Prop->SetPropertyValue( (void*)(Property->ContainerPtrToValuePtr<uint16>( &PacketHeader )), Value );
			}
			else if (Property->IsA( FIntProperty::StaticClass( ) ))
			{
				int32 Value = GetInt32( RawData, Offset );
				FIntProperty* IntProp = CastField<FIntProperty>( Property );
				IntProp->SetPropertyValue( (void*)(Property->ContainerPtrToValuePtr<int32>( &PacketHeader )), Value );
			}
			else if (Property->IsA( FUInt32Property::StaticClass( ) ))
			{
				uint32 Value = GetUInt32( RawData, Offset );
				FUInt32Property* UIntProp = CastField<FUInt32Property>( Property );
				UIntProp->SetPropertyValue( (void*)(Property->ContainerPtrToValuePtr<uint32>( &PacketHeader )), Value );
			}
			else if (Property->IsA( FFloatProperty::StaticClass( ) ))
			{
				float Value = GetFloat( RawData, Offset );
				FFloatProperty* FloatProp = CastField<FFloatProperty>( Property );
				FloatProp->SetPropertyValue( (void*)(Property->ContainerPtrToValuePtr<float>( &PacketHeader )), Value );
			}
			else if (Property->IsA( FNameProperty::StaticClass( ) ))
			{
				FString PropName = Property->GetName( );
				if (FString( TEXT( "AvatarName" ) ) == PropName)
				{
					PacketHeader.AvatarName = FName( *GetString( RawData, Offset, FHi5PacketHeader::AvatarNameSize ) );	// temp: name length is 32 bytes.  We need to send name length within the protocol which can bear changes from server side.
				}
			}
		}

		return Offset == FHi5PacketHeader::HeaderSize;
	}
}
