// Copyright 2015 Noitom Technology Co., Ltd. All Rights Reserved.

#include "MotionCaptureManager.h"
#include "MotionCaptureSpawner.h"
#include "EventManager.h"
#include "EventType.h"
#include "AnimationDataManager.h"
#include "MotionCaptureNetwork.h"
#include "MotionCaptureComponent.h"
#include "EngineUtils.h"
#include "Engine/Texture2D.h"
#include "UObject/ConstructorHelpers.h"

#if WITH_EDITOR
#include "Components/BillboardComponent.h"
#endif

AMotionCaptureManager* AMotionCaptureManager::Instance = nullptr;

AMotionCaptureManager::AMotionCaptureManager(const FObjectInitializer& ObjectInitializer)
	: Super(ObjectInitializer)
{
	Instance = this;
	USceneComponent* SceneComponent = CreateDefaultSubobject<USceneComponent>(TEXT("SceneComp"));
	RootComponent = SceneComponent;
	
#if WITH_EDITOR
	SpriteComponent = CreateEditorOnlyDefaultSubobject<UBillboardComponent>(TEXT("Sprite"));

	if (!IsRunningCommandlet())
	{
		// Structure to hold one-time initialization
		struct FConstructorStatics
		{
			ConstructorHelpers::FObjectFinderOptional<UTexture2D> StartPointIconObject;
			FName ID_StartPoint;
			FText NAME_StartPoint;
			FConstructorStatics()
				: StartPointIconObject(TEXT("/MotionCapture/Icons/ManagerIcon"))
				, ID_StartPoint(TEXT("Misc"))
				, NAME_StartPoint(NSLOCTEXT("SpriteCategory", "Misc", "Misc"))
			{
			}
		};
		static FConstructorStatics ConstructorStatics;

		if (SpriteComponent)
		{
			SpriteComponent->Sprite = ConstructorStatics.StartPointIconObject.Get();
#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION < 24
			SpriteComponent->RelativeScale3D = FVector(0.55f, 0.55f, 0.55f);
#else
			SpriteComponent->SetRelativeScale3D( FVector( 0.55f, 0.55f, 0.55f ) );
#endif
			SpriteComponent->SpriteInfo.Category = ConstructorStatics.ID_StartPoint;
			SpriteComponent->SpriteInfo.DisplayName = ConstructorStatics.NAME_StartPoint;
			SpriteComponent->bIsScreenSizeScaled = true;
			SpriteComponent->SetupAttachment(SceneComponent);
			SpriteComponent->SetRelativeLocation(FVector(.0f, .0f, 10.0f));
		}
	}
#endif

	PrimaryActorTick.bCanEverTick = true;
}

AMotionCaptureManager* AMotionCaptureManager::GetInstance()
{
	return Instance;
}

bool AMotionCaptureManager::Connect(const FString& sIP, int32 iPort)
{
	if (nullptr != Network)
	{
		if (!Network->Connect(sIP,iPort))
		{
			PN_LOG(Error, TEXT("Bound to %s:%s failed."), *IP, *(FString::FromInt(Port)));
			return false;
		}
		else
		{
			PN_LOG(Log, TEXT("Bound to %s:%s..."), *IP, *(FString::FromInt(Port)));
			return true;
		}
	}

	return false;
}

void AMotionCaptureManager::Disconnect()
{
	if (Network != nullptr)
	{
		Network->DisConnect();
		Network = nullptr;
		PN_LOG(Log, TEXT("Network DisConnect."));
	}
}

void AMotionCaptureManager::BeginPlay()
{
	if (Network != nullptr)
	{
		Network->DisConnect();
	}
	else
	{
		Network = NewObject<UMotionCaptureNetwork>(this);
	}
	
	// Setup PNNetwork
	if (nullptr != Network)
	{
		if (!Network->Connect())
		{
			PN_LOG(Error, TEXT("Bound to %s:%s failed."), *IP, *(FString::FromInt(Port)));
		}
		else
		{
			PN_LOG(Log, TEXT("Bound to %s:%s..."), *IP, *(FString::FromInt(Port)));
		}
	}

	// Collect PNComponent
	for (TActorIterator<AMotionCaptureSpawner> It(GetWorld()); It; ++It)
	{
		AMotionCaptureSpawner* Spawner = Cast<AMotionCaptureSpawner>(*It);
		if (nullptr != Spawner)
		{
			It->PNManager = this;
			SpawnerInstances.Add(*It);
		}
	}

	Super::BeginPlay();
}

void AMotionCaptureManager::BeginDestroy()
{
	if (Network != nullptr)
	{
		Network->DisConnect();
		Network = nullptr;
		PN_LOG(Log, TEXT("Network DisConnect."));
	}
	
	//SpawnerInstances.Empty();
	FAnimationDataManager::GetInstance()->RemoveAllAnimationData();
	FEventManager::GetInstance()->ClearEventQueue();
	
	return Super::BeginDestroy();
}

void AMotionCaptureManager::Tick(float DeltaSeconds)
{
	FEventManager::GetInstance()->Tick();
}

bool AMotionCaptureManager::FindSpawner(int32 AvatarID, const FName& AvatarName, FVector& InitPos, FQuat& InitQua, UClass** ActorClass) const
{
	bool bMatch = false;
	for (auto Spawner : SpawnerInstances)
	{
		if (!Spawner.IsValid())
		{
			continue;
		}

#if WITH_AVATARID
		if (Spawner->AvatarID == AvatarID)
		{
			bMatch = true;
		}
#else
		if (Spawner->AvatarName == AvatarName)
		{
			bMatch = true;
		}
#endif

		if (bMatch)
		{
			InitPos = Spawner->GetActorLocation();
			InitQua = Spawner->GetActorQuat();
			*ActorClass = Spawner->DynamicActorClass;
			SpawnerInstances.Remove(Spawner);
			break;
		}
	}

	return bMatch;
}

void AMotionCaptureManager::CreatePNActor(const FEventBase* Event) const
{
	const FCreateNewEvent* TargetEvent = static_cast<const FCreateNewEvent*>(Event);
	if (TargetEvent && GetWorld())
	{
		ULevel* InLevel = GetWorld()->GetCurrentLevel();
		FActorSpawnParameters SpawnInfo;
		SpawnInfo.OverrideLevel = InLevel;
		SpawnInfo.ObjectFlags = RF_Transient;

		while (true)
		{
			FVector InitPos = TargetEvent->InitPos;
			FQuat InitQua = TargetEvent->InitQua;
			UClass* ActorClass = nullptr;
			bool bFound = FindSpawner(TargetEvent->AvatarID, TargetEvent->AvatarName, InitPos, InitQua, &ActorClass);
			if (!bFound)
			{
				break;
			}

			if (nullptr != ActorClass)
			{
				FRotator InitRot = InitQua.Rotator();
				SpawnInfo.Name = MakeUniqueObjectName(InLevel, ActorClass, TargetEvent->AvatarName);
				AActor* NewActor = GetWorld()->SpawnActor(ActorClass, &InitPos, &InitRot, SpawnInfo);
				if (!NewActor)
				{
					PN_LOG(Error, TEXT("CreatePNActor Failed!  Spawn Actor failed."));
					continue;
				}
				UMotionCaptureComponent* PNComp = NewActor->FindComponentByClass<UMotionCaptureComponent>();
				if (!PNComp)
				{
					PN_LOG(Error, TEXT("CreatePNActor Failed!  Actor class is invalid that doesn't contain any MotionCaptureComponent."));
					continue;
				}
#if WITH_AVATARID
				PNComp->SetAvatarID(TargetEvent->AvatarID);
#else
				PNComp->SetAvatarName(TargetEvent->AvatarName);
#endif
			}
			else
			{
				PN_LOG(Error, TEXT("The world or actor class is invalid for creat new Actor."));
			}
		}
	}
}

#if WITH_EDITOR
/** Returns SpriteComponent subobject **/
UBillboardComponent* AMotionCaptureManager::GetSpriteComponent() const { return SpriteComponent; }
#endif
