// Copyright 2015 Noitom Technology Co., Ltd. All Rights Reserved.

#include "IMotionCaptureModule.h"
#include "Interfaces/IPluginManager.h"
#include "MotionCaptureFunctionLibrary.h"
#include "HI5Glove.h"
#include "Hi5_Manager_Thread.h"
#include "Misc/MessageDialog.h"
#include "HAL/PlatformProcess.h"

#define LOCTEXT_NAMESPACE "FMotionCapture"

class FMotionCapture : public IMotionCapture
{
	/** IModuleInterface implementation */
	virtual void StartupModule() override;
	virtual void ShutdownModule() override;
	virtual void OnReceiveGloveData(const FString& AvartarName, const TArray<uint8>& Data) override;
	virtual bool GetLocalGloveData (EGloveMod Glove, const FString &AvatarName, TArray<uint8>& Data) override;
	virtual void CacheHybridDataCaliProgress (int32 Pose, int32 percent) override;
	virtual void GetCachedHybridDataCaliProgress (ECalibrationPose& Pose, int32& Percent) override;
	virtual bool GetParsedLocalJointsData (EGloveMod Glove, TArray<FVector>& Positions, TArray<FRotator>& Orientations) override;
	virtual bool GetLocalJointData (EGloveMod Glove, EMCBones::Type Bone, FVector& Position, FRotator& Orientation) override;

private:
	HI5Device::CalibrationPercent m_stCaliPercent;

	std::vector<void*> Hi5CppSDKLibraryHandles;
};


void FMotionCapture::StartupModule()
{
	// This code will execute after your module is loaded into memory (but after global variables are initialized, of course.)

	PN_LOG(Display, TEXT("MotionCapture plugin startup."));

	// Get the base directory of this plugin
	FString BaseDir = IPluginManager::Get ().FindPlugin ("MotionCapture")->GetBaseDir ();

	// Add on the relative location of the third party dll and load it
	FString LibraryDirectory = FPaths::Combine (*BaseDir, TEXT ("Source/HI5Glove_SDK/bin"));
	FPlatformProcess::AddDllDirectory( *LibraryDirectory );

	bool Hi5SDK_Missing = false;
	if (void* LibraryHandle = FPlatformProcess::GetDllHandle( *FPaths::Combine( *LibraryDirectory, TEXT( "Hi5Glove.dll" ) ) ))
	{
		Hi5CppSDKLibraryHandles.push_back( LibraryHandle );
	}
	else
	{
		Hi5SDK_Missing = true;
	}

	if (Hi5SDK_Missing)
	{
		FMessageDialog::Open (EAppMsgType::Ok, LOCTEXT ("ThirdPartyLibraryError", "Failed to load Hi5Cpp SDK library"));
	}
	
}


void FMotionCapture::ShutdownModule()
{
	// This function may be called during shutdown to clean up your module.  For modules that support dynamic reloading,
	// we call this function before unloading the module.

	// Free the dll handle
	for (std::vector<void*>::reverse_iterator riter = Hi5CppSDKLibraryHandles.rbegin (); riter != Hi5CppSDKLibraryHandles.rend (); ++riter) {
		FPlatformProcess::FreeDllHandle (*riter);
	}
	Hi5CppSDKLibraryHandles.clear ();

	PN_LOG(Display, TEXT("MotionCapture plugin shutdown."));
}

void FMotionCapture::OnReceiveGloveData(const FString& AvartarName, const TArray<uint8>& Data)
{
	UMotionCaptureFunctionLibrary::OnReceiveGloveData(AvartarName, Data);
}

bool FMotionCapture::GetLocalGloveData (EGloveMod Glove, const FString &AvatarName, TArray<uint8>& Data)
{
	return UMotionCaptureFunctionLibrary::GetLocalGloveData (Glove, AvatarName, Data);
}

void FMotionCapture::CacheHybridDataCaliProgress (int32 Pose, int32 percent)
{
	if (Pose == CG_CalibBPosePer)
	{
		m_stCaliPercent.pos = CalibrationPose::GCP_BPose;
	}
	else if (Pose == CG_CalibPPosePer)
	{
		m_stCaliPercent.pos = CalibrationPose::GCP_PPose;
	}

	m_stCaliPercent.percent = percent;
}

void FMotionCapture::GetCachedHybridDataCaliProgress (ECalibrationPose& Pose, int32& Percent)
{
	Pose = (ECalibrationPose)(int)m_stCaliPercent.pos;
	Percent = m_stCaliPercent.percent;
}

bool FMotionCapture::GetParsedLocalJointsData (EGloveMod Glove, TArray<FVector>& Positions, TArray<FRotator>& Orientations)
{
	return UMotionCaptureFunctionLibrary::GetParsedLocalJointsData (Glove, Positions, Orientations);
}

bool FMotionCapture::GetLocalJointData (EGloveMod Glove, EMCBones::Type Bone, FVector& Position, FRotator& Orientation)
{
	return UMotionCaptureFunctionLibrary::GetLocalJointData (Glove, Bone, Position, Orientation);
}

#undef LOCTEXT_NAMESPACE

IMPLEMENT_MODULE(FMotionCapture, MotionCaptureRuntime)

