// Copyright 2015 Noitom Technology Co., Ltd. All Rights Reserved.

#pragma once

#include "CoreMinimal.h"

namespace Hi5
{
	/**
	* Object synchronizing read access (from multiple thread) and write access (only from one thread at the same time).
	**/
	class FPNMultiReaderSingleWriter
	{
	public:
		/** Protect data from modification while reading. */
		void LockRead( );

		/** Ends protecting data from modification while reading. */
		void UnlockRead( );

		/** Protect data from reading while modifying. */
		void LockWrite( );

		/** Ends protecting data from reading while modifying. */
		void UnlockWrite( );

	private:
		struct FMRSWCriticalSection
		{
			FMRSWCriticalSection( )
				: Action( 0 )
				, ReadCounter( 0 )
				, WriteCounter( 0 )
			{ }

			int32 Action;
			FThreadSafeCounter ReadCounter;
			FThreadSafeCounter WriteCounter;
		};

		FMRSWCriticalSection CriticalSection;

		static const int32 WritingAction = -1;
		static const int32 NoAction = 0;
		static const int32 ReadingAction = 1;

		bool CanRead( )
		{
			return FPlatformAtomics::InterlockedCompareExchange( &CriticalSection.Action, ReadingAction, NoAction ) == ReadingAction;
		}

		bool CanWrite( )
		{
			return FPlatformAtomics::InterlockedCompareExchange( &CriticalSection.Action, WritingAction, NoAction ) == NoAction;
		}
	};


	class FPNReadScopeLock
	{
	public:
		/**
		* Constructor that performs a lock on the synchronization object
		*
		* @param InSynchObject The synchronization object to manage
		*/
		FPNReadScopeLock( FPNMultiReaderSingleWriter* InSynchObject )
			: SynchObject( InSynchObject )
		{
			check( SynchObject );
			SynchObject->LockRead( );
		}

		/** Destructor that performs a release on the synchronization object. */
		~FPNReadScopeLock( )
		{
			check( SynchObject );
			SynchObject->UnlockRead( );
		}

		/** Default constructor (delete on purpose). */
		FPNReadScopeLock( ) = delete;

		/** Copy constructor( delete on purpose). */
		FPNReadScopeLock( const FPNReadScopeLock& InScopeLock ) = delete;

		/** Assignment operator (delete on purpose). */
		FPNReadScopeLock& operator=( const FPNReadScopeLock& InScopeLock ) = delete;

	private:
		// Holds the synchronization object to aggregate and scope manage.
		FPNMultiReaderSingleWriter* SynchObject;
	};


	class FPNWriteScopeLock
	{
	public:
		/**
		* Constructor that performs a lock on the synchronization object
		*
		* @param InSynchObject The synchronization object to manage
		*/
		FPNWriteScopeLock( FPNMultiReaderSingleWriter* InSynchObject )
			: SynchObject( InSynchObject )
		{
			check( SynchObject );
			SynchObject->LockWrite( );
		}

		/** Destructor that performs a release on the synchronization object. */
		~FPNWriteScopeLock( )
		{
			check( SynchObject );
			SynchObject->UnlockWrite( );
		}

		/** Default constructor (delete on purpose). */
		FPNWriteScopeLock( ) = delete;

		/** Copy constructor( delete on purpose). */
		FPNWriteScopeLock( const FPNWriteScopeLock& InScopeLock ) = delete;

		/** Assignment operator (delete on purpose). */
		FPNWriteScopeLock& operator=( const FPNWriteScopeLock& InScopeLock ) = delete;

	private:
		// Holds the synchronization object to aggregate and scope manage.
		FPNMultiReaderSingleWriter* SynchObject;
	};
}

